from pweb_form_rest.crud.pweb_crud_common import PWebCRUDCommon
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_orm import PWebBaseModel


class RESTDataCRUD(PWebCRUDCommon):

    def __init__(self, model: PWebBaseModel):
        self.model = model

    def create(self, request_dto: PWebDataDTO, response_dto: PWebDataDTO = None, response_message: str = "Successfully created!", data: dict = None):
        if not data:
            data = self.get_json_data(request_dto)
        model = self.save(data=data, request_dto=request_dto)
        return self.message_or_data_response(model, response_dto, response_message)

    def details(self, model_id: int, response_dto: PWebDataDTO, query=None):
        existing_model = self.get_by_id(model_id, exception=True, query=query)
        return self.response_maker.data_response(existing_model, response_dto)

    def update(self, request_dto: PWebDataDTO, response_dto: PWebDataDTO = None, response_message: str = "Successfully updated!", existing_model=None, data: dict = None, query=None):
        if not data:
            data = self.get_json_data(request_dto)
        model = self.edit(model_id=data['id'], data=data, request_dto=request_dto, existing_model=existing_model, query=query)
        return self.message_or_data_response(model, response_dto, response_message)

    def delete(self, model_id: int, response_message: str = "Successfully deleted!", query=None):
        self.soft_remove(model_id=model_id, query=query)
        return self.response_maker.success_message(response_message)

    def hard_delete(self, model_id: int, response_message: str = "Successfully deleted!", query=None):
        self.remove(model_id, query=query)
        return self.response_maker.success_message(response_message)

    def paginated_list(self, response_dto: PWebDataDTO, query=None, search_fields: list = None, sort_field=None, sort_order=None, item_per_page=None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, item_per_page=item_per_page)
        return self.response_maker.paginate_response(data_list, response_dto)

    def list(self, response_dto: PWebDataDTO, query=None, search_fields: list = None, sort_field=None, sort_order=None):
        data_list = self.read_all(query=query, search_fields=search_fields, sort_field=sort_field, sort_order=sort_order, enable_pagination=False)
        return self.response_maker.list_data_type_response(data_list, response_dto=response_dto)



from functools import wraps
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_constant import SwaggerDefinitionType, HTTPContentType, SwaggerCommon, HTTPMethod, \
    SwaggerDataType
from pweb_form_rest.swagger.pweb_swagger_param_def import SwaggerParamDef


def add_swagger_endpoint(
        request_obj=None, request_list=None, response_obj=None,
        response_list=None, query_params: list = None,
        url_params: list = None, tag: str = None, method: str = None,
        def_type: str = SwaggerDefinitionType.NONE, http_response_code: int = 200,
        response_content_type: str = HTTPContentType.APPLICATION_JSON,
        request_content_type: str = HTTPContentType.APPLICATION_JSON,
        pweb_message_response: bool = False, pweb_error_details_response: bool = False):
    def decorator(function):
        function.__pweb_swagger__ = SwaggerCommon.PWEB_SWAGGER

        @wraps(function)
        def pweb_swagger_def(*args, **kwargs):
            if 'pass_definition' in kwargs and kwargs['pass_definition']:
                definition = SwaggerParamDef()
                definition.request_obj = request_obj
                definition.request_list = request_list
                definition.response_obj = response_obj
                definition.response_list = response_list
                definition.query_params = query_params
                definition.url_params = url_params
                definition.http_response_code = http_response_code
                definition.response_content_type = response_content_type
                definition.request_content_type = request_content_type
                definition.pweb_error_details_response = pweb_error_details_response
                definition.tag = tag
                definition.method = method
                definition.def_type = def_type
                definition.pweb_message_response = pweb_message_response
                return definition
            return function(*args, **kwargs)

        return pweb_swagger_def

    return decorator


def pweb_endpoint(
        request_obj=None, request_list=None, response_obj=None, method: str = None, response_list=None,
        query_params: list = None, url_params: list = None, tag: str = None, def_type: str = SwaggerDefinitionType.NONE,
        http_response_code: int = 200, response_content_type: str = HTTPContentType.APPLICATION_JSON,
        request_content_type: str = HTTPContentType.APPLICATION_JSON, pweb_message_response: bool = False,
        pweb_error_details_response: bool = False):
    return add_swagger_endpoint(
        request_obj=request_obj, request_list=request_list, response_obj=response_obj, response_list=response_list,
        method=method, query_params=query_params, url_params=url_params, def_type=def_type,
        tag=tag, pweb_message_response=pweb_message_response, pweb_error_details_response=pweb_error_details_response,
        http_response_code=http_response_code, response_content_type=response_content_type,
        request_content_type=request_content_type
    )


def pweb_upload_endpoint(
        request_obj=None, request_list=None, response_obj=None,
        response_list=None, query_params: list = None,
        url_params: list = None, tag: str = None,
        http_response_code: int = 200, response_content_type: str = HTTPContentType.APPLICATION_JSON,
        pweb_message_response: bool = False, pweb_error_details_response: bool = False):
    return add_swagger_endpoint(
        def_type=SwaggerDefinitionType.FILE_UPLOAD,
        request_obj=request_obj, request_list=request_list, response_obj=response_obj, response_list=response_list,
        method=HTTPMethod.POST, query_params=query_params, url_params=url_params,
        tag=tag, pweb_message_response=pweb_message_response,
        http_response_code=http_response_code, response_content_type=response_content_type,
        request_content_type=HTTPContentType.MULTIPART_FORM_DATA,
        pweb_error_details_response=pweb_error_details_response
    )


def pweb_paginate_endpoint(
        response_obj=None, query_params: list = None,
        url_params: list = None, tag: str = None, method: str = HTTPMethod.GET,
        pagination: bool = True, sorting: bool = True, search: bool = True,
        http_response_code: int = 200, response_content_type: str = HTTPContentType.APPLICATION_JSON,
        pweb_message_response: bool = False):
    if not query_params:
        query_params = []

    if pagination:
        query_params.append((PWebFRConfig.PAGE_PARAM_NAME, SwaggerDataType.integer))
        query_params.append((PWebFRConfig.ITEM_PER_PAGE_PARAM_NAME, SwaggerDataType.integer))

    if sorting:
        query_params.append((PWebFRConfig.SORT_FIELD_PARAM_NAME, SwaggerDataType.string))
        query_params.append((PWebFRConfig.SORT_ORDER_PARAM_NAME, SwaggerDataType.string))

    if search:
        query_params.append((PWebFRConfig.SEARCH_FIELD_PARAM_NAME, SwaggerDataType.string))

    return add_swagger_endpoint(
        method=method, query_params=query_params, url_params=url_params,
        tag=tag, pweb_message_response=pweb_message_response, response_list=response_obj,
        http_response_code=http_response_code, response_content_type=response_content_type
    )



from flask import Blueprint, render_template, request
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.swagger.pweb_sd_processor import PWebSDProcessor
from pweb_form_rest.swagger.pweb_swagger_generator import PWebSwaggerGenerator


class PWebSwaggerUI:
    _pweb_app = None

    def register(self, pweb_app):
        self._pweb_app = pweb_app
        self.init_swagger_blueprint()

    def init_swagger_blueprint(self):
        if PWebFRConfig.ENABLE_SWAGGER_UI and self._pweb_app:
            blueprint = Blueprint(
                "PWebSwaggerSystem",
                __name__,
                template_folder="template-assets/templates",
                static_folder="template-assets/assets",
                static_url_path=PWebFRConfig.SWAGGER_UI_ASSETS_URL
            )
            blueprint.add_url_rule(PWebFRConfig.SWAGGER_JSON_URL, "pweb-swagger-json", self.swagger_json)
            blueprint.add_url_rule(PWebFRConfig.SWAGGER_UI_URL, "pweb-swagger-ui", self.swagger_ui)
            self._pweb_app.register_blueprint(blueprint)

    def swagger_ui(self):
        auth = self.check_auth()
        if auth:
            return auth
        return render_template('pweb-swagger-ui.html', config=PWebFRConfig)

    def swagger_json(self):
        auth = self.check_auth()
        if auth:
            return auth
        decorator_processor = PWebSDProcessor(self._pweb_app)
        action_definitions = decorator_processor.get_action_definitions()
        pweb_swagger_generator = PWebSwaggerGenerator()
        pweb_swagger_generator.process_action_definitions(action_definitions)
        return pweb_swagger_generator.get_swagger_spec()

    def check_auth(self):
        if PWebFRConfig.ENABLE_SWAGGER_AUTH:
            auth = request.authorization
            if not (auth and auth.username == PWebFRConfig.SWAGGER_AUTH_USERNAME and auth.password == PWebFRConfig.SWAGGER_AUTH_PASSWORD):
                return ('You are not authorize to access the URL.', 401, {
                    'WWW-Authenticate': 'Basic realm="Login Required"'
                })
        return None



from marshmallow import Schema, fields
from pweb_form_rest.common.pweb_fr_constant import SwaggerCommon
from pweb_form_rest.swagger.pweb_swagger_param_def import SwaggerParamDef


class PWebSwaggerSchema:
    IN_PATH = "path"
    IN_QUERY = "query"
    OBJ = "object"
    ARRAY = "array"
    JWT_SCHEME = {"type": "http", "scheme": "bearer", "bearerFormat": "JWT"}

    @staticmethod
    def get_url_param_schema(place, name, data_type, required=False):
        return {
            "name": name,
            "in": place,
            "schema": {
                "type": data_type
            },
            "required": required
        }

    @staticmethod
    def get_component_schemas_link(name):
        return "#/components/schemas/" + name

    @staticmethod
    def get_schema_type(many):
        schema_type = None
        if many:
            schema_type = PWebSwaggerSchema.ARRAY
        return schema_type

    @staticmethod
    def get_schema_def_and_ref(ref, schema_type=None):
        schema = {"type": schema_type}
        if schema_type == PWebSwaggerSchema.ARRAY:
            schema["items"] = {
                "$ref": PWebSwaggerSchema.get_component_schemas_link(ref)
            }
            return schema
        return {"$ref": PWebSwaggerSchema.get_component_schemas_link(ref)}

    @staticmethod
    def get_request_body(definition: SwaggerParamDef, many=False):
        schema_type = PWebSwaggerSchema.get_schema_type(many)
        return {
            "required": True,
            "content": {
                definition.request_content_type: {
                    "schema": PWebSwaggerSchema.get_schema_def_and_ref(definition.request_schema_key, schema_type)
                }
            }
        }

    @staticmethod
    def get_response_body(definition: SwaggerParamDef, many=False):
        schema_type = PWebSwaggerSchema.get_schema_type(many)
        schema = PWebSwaggerSchema.get_schema_def_and_ref(definition.response_schema_key, schema_type)

        any_of_response = {"anyOf": []}
        if definition.response_obj or definition.response_list:
            any_of_response["anyOf"].append(PWebSwaggerSchema.get_schema_def_and_ref(definition.response_schema_key))
        if definition.pweb_message_response:
            any_of_response["anyOf"].append(PWebSwaggerSchema.get_schema_def_and_ref(SwaggerCommon.MESSAGE_RESPONSE))
        if definition.pweb_error_details_response:
            any_of_response["anyOf"].append(PWebSwaggerSchema.get_schema_def_and_ref(SwaggerCommon.ERROR_DETAILS_RESPONSE))
        schema = any_of_response

        return {
            definition.http_response_code: {
                "content": {
                    definition.response_content_type: {
                        "schema": schema
                    }
                }
            }
        }

    @staticmethod
    def pweb_api_data_schema(data, many=False):
        return Schema.from_dict({
            "data": fields.Nested(data, many=many)
        })

    @staticmethod
    def pweb_api_response_base_schema_map():
        return {
            "status": fields.String(),
            "code": fields.String(),
        }

    @staticmethod
    def pweb_api_response_data_schema(data, many=False, response_map=False):
        schema_map = PWebSwaggerSchema.pweb_api_response_base_schema_map()
        if data:
            schema_map["data"] = fields.Nested(data, many=many)
        if response_map:
            return schema_map
        return Schema.from_dict(schema_map)

    @staticmethod
    def pweb_api_paginate_response_schema(data):
        schema_map = PWebSwaggerSchema.pweb_api_response_data_schema(data, many=True, response_map=True)
        pagination_schema = Schema.from_dict({
            "page": fields.Integer(),
            "itemPerPage": fields.Integer(),
            "total": fields.Integer(),
            "totalPage": fields.Integer(),
        })
        schema_map["pagination"] = fields.Nested(pagination_schema)
        return Schema.from_dict(schema_map)

    @staticmethod
    def pweb_api_message_response_schema():
        schema_map = PWebSwaggerSchema.pweb_api_response_base_schema_map()
        schema_map["message"] = fields.String()
        return Schema.from_dict(schema_map)

    @staticmethod
    def pweb_api_error_response_schema():
        schema_map = PWebSwaggerSchema.pweb_api_response_base_schema_map()
        schema_map["message"] = fields.String()
        schema_map["error"] = fields.Dict(keys=fields.String(), values=fields.String())
        return Schema.from_dict(schema_map)



from pweb_form_rest.schema.pweb_rest_schema import APIBase, APIMessageResponse, APIErrorResponse, APIDataResponse, \
    APIDataListResponse, APIPaginateResponse


class PWebBaseResponse:
    status: str = None
    code: str = None
    httpCode: int = None
    message: str = None

    def make_dict(self, data, api_base: APIBase, many=False):
        return api_base.dumps(data, many=many)


class PWebMessageResponse(PWebBaseResponse):
    message: str = None

    def to_dict(self):
        return self.make_dict(self, APIMessageResponse())


class PWebErrorResponse(PWebMessageResponse):
    error: dict = None

    def to_dict(self):
        return self.make_dict(self, APIErrorResponse())


class PWebDataResponse(PWebBaseResponse):
    _schema = APIDataResponse()
    data: any

    def __init__(self):
        self.data = None

    def add_data(self, model, schema: APIBase, many=False):
        if many:
            self._schema = APIDataListResponse()
        else:
            self._schema = APIDataResponse()

        if model and schema:
            self.add_only_data(model, schema, many)

    def add_only_data(self, model, schema: APIBase, many=False):
        if model:
            self.data = schema.dump(model, many=many)

    def to_dict(self):
        return self.make_dict(self, self._schema)


class PWebPagination(object):
    page: int = None
    itemPerPage: int = None
    total: int = None
    totalPage: int = None


class PWebPaginateResponse(PWebDataResponse):
    _schema = APIPaginateResponse()
    pagination: PWebPagination = None

    def to_dict(self):
        return self.make_dict(self, self._schema)



from ppy_common import PPyCException
from pweb_form_rest.data.pweb_response_data import PWebMessageResponse, PWebErrorResponse
from pweb_form_rest.data.pweb_response_status import PWebResponseCode, PWebHTTPCode


class FormRESTException(PPyCException):
    messageResponse: PWebMessageResponse
    errorResponse: PWebErrorResponse

    def __init__(self, message=None, exception_type: str = None):
        self._set_super(message, exception_type)

    def _set_super(self, message=None, exception_type: str = None):
        super().__init__(message, exception_type)

    def error_message_exception(self, message: str, code=PWebResponseCode.error, http_code=PWebHTTPCode.OK):
        self._set_super(message)
        response = PWebMessageResponse()
        response.message = message
        response.code = code
        response.status = PWebResponseCode.error
        response.http_code = http_code
        self.messageResponse = response
        return self

    def error_details_exception(self, message: str, details: dict, code=PWebResponseCode.error, http_code=PWebHTTPCode.OK):
        self._set_super(message)
        response = PWebErrorResponse()
        response.message = message
        response.code = code
        response.status = PWebResponseCode.error
        response.http_code = http_code
        response.error = details
        self.messageResponse = response
        return self

    def process_validation_exception(self, errors: dict, message: str):
        message_dict: dict = {}
        for message in errors:
            error_text = ""
            for text in errors[message]:
                error_text += str(text) + " "
            message_dict[message] = error_text
        return self.error_details_exception(message=message, details=message_dict)


form_rest_exception = FormRESTException()



from marshmallow import ValidationError
from pweb_form_rest.common.pweb_fr_exception import FormRESTException
from pweb_form_rest.crud.pweb_crud import PWebCRUD
from pweb_form_rest.crud.pweb_request_data import RequestData
from pweb_form_rest.form.pweb_form_definition import FormDefinition
from pweb_form_rest.schema.pweb_rest_schema import PWebOrmDTO


class PWebFormData:
    request_data: RequestData = RequestData()
    pweb_crud: PWebCRUD = PWebCRUD()

    def is_post_data(self) -> bool:
        return self.request_data.is_post()

    def is_get_data(self) -> bool:
        return self.request_data.is_get()

    def handle_validation_exception(self, exception, definition: FormDefinition):
        errors = {}
        if exception and exception.messages_dict and isinstance(exception.messages_dict, dict):
            for name, error in exception.messages_dict.items():
                errors[name] = ', '.join(error)
        definition.set_field_errors(errors)

    def handle_form_rest_exception(self, exception, definition: FormDefinition):
        if exception.messageResponse and hasattr(exception.messageResponse, "error") and exception.messageResponse.error:
            definition.set_field_errors(exception.messageResponse.error)

    def is_valid_data(self, form: PWebOrmDTO, definition: FormDefinition, form_data=None) -> bool:
        try:
            if not form_data:
                form_data = definition.get_request_data()

            if form_data:
                definition.set_field_value(field_and_value=form_data)

            if definition.is_validation_error:
                return False

            self.pweb_crud.get_form_data(data_dto=form, form_data=form_data)
            return True
        except ValidationError as e:
            self.handle_validation_exception(exception=e, definition=definition)
        except FormRESTException as e:
            self.handle_form_rest_exception(exception=e, definition=definition)
        return False



from pweb_form_rest.schema.pweb_rest_schema import APIBase
from pweb_form_rest.common.pweb_custom_field import FileField, EnumField, CustomNestedField
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.swagger.pweb_swagger_decorator import pweb_endpoint, pweb_paginate_endpoint, pweb_upload_endpoint
from pweb_form_rest.pweb_fr import PWebFR
from pweb_form_rest.crud.pweb_rest_data_crud import RESTDataCRUD
from pweb_form_rest.crud.pweb_file_data_crud import FileDataCRUD
from pweb_form_rest.ui.pweb_ui_helper import PWebSSRUIHelper, ssr_ui_render
from pweb_form_rest.form.pweb_form_field import FormField
from pweb_form_rest.form.pweb_form import PWebForm
from pweb_form_rest.common.pweb_fr_exception import form_rest_exception



class SwaggerDefinitionType(object):
    NONE = None
    PAGINATION = "PAGINATION"
    FILE_UPLOAD = "FILE_UPLOAD"
    FORM_DATA = "FORM_DATA"


class SwaggerCommon(object):
    PWEB_SWAGGER = "PWEB_SWAGGER"
    MESSAGE_RESPONSE = "message_response"
    ERROR_DETAILS_RESPONSE = "error_response"


class SwaggerDataType(object):
    integer = "integer"
    number = "number"
    string = "string"


class HTTPContentType(object):
    APPLICATION_JSON = "application/json"
    MULTIPART_FORM_DATA = "multipart/form-data"
    FORM_DATA = "application/x-www-form-urlencoded"


class HTTPMethod(object):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"



import re
from urllib.parse import urlparse, parse_qsl, urlencode
from flask import request
from werkzeug.datastructures import ImmutableMultiDict
from pweb_form_rest.data.pweb_request_info import PWebRequestInfo


class RequestData:

    def _parse_value(self, value):
        if value == "null" or value == "None":
            return None
        return value

    def _form_data_dict_to_dict(self, _input: ImmutableMultiDict, default=None):
        if not _input:
            return default
        requested_data = _input.to_dict(flat=False)
        response = {}
        for data in requested_data:
            if len(requested_data[data]) == 1:
                response[data] = self._parse_value(requested_data[data][0])
            else:
                response[data] = self._parse_value(requested_data[data])
        return response

    def json_data(self, default=None):
        json = request.get_json()
        if json:
            return json
        return default

    def form_data(self, default=None):
        data = request.form
        if data:
            return self._form_data_dict_to_dict(data, default)
        return default

    def file_data(self, default=None):
        files = request.files
        if files:
            return self._form_data_dict_to_dict(files, default)
        return default

    def form_and_file_data(self, default=None):
        form_data = self.form_data(default={})
        file_data = self.file_data(default={})
        form_data.update(file_data)
        if form_data:
            return form_data
        return default

    def query_args(self, default=None):
        data = request.args
        if data:
            return data
        return default

    def get_query_args_value(self, key: str, default=None, type=None):
        data = self.query_args(default=None)
        if not data:
            return default
        if key in data:
            value = data.getlist(key, type)
            if len(value) == 1:
                return value[0]
            return value
        return default

    def request_method(self):
        return request.method

    def is_post(self):
        return 'POST' == self.request_method()

    def is_get(self):
        return 'GET' == self.request_method()

    def current_url(self):
        return request.referrer

    def get_header(self, name: str, default=None):
        return request.headers.get(name, default)

    def get_auth_header(self):
        header = self.get_header("Authorization")
        if not header:
            header = self.get_header("authorization")
        return header

    def get_bearer_token(self):
        authorization_header = self.get_auth_header()
        if not authorization_header:
            return None
        group = re.match("^Bearer\\s+(.*)", authorization_header)
        if group:
            return group.group(1)
        return None

    def get_url_info(self) -> PWebRequestInfo:
        url_dictionary = PWebRequestInfo()
        if request and request.url:
            if request.path:
                url_dictionary.relativeURL = str(request.path)
            url_dictionary.relativeURLWithParam = str(request.full_path)
            url_dictionary.hostWithPort = str(request.host)
            url_dictionary.method = str(request.method)
            url_dictionary.urlRule = str(request.url_rule)
            url_dictionary.baseURL = str(request.host_url).strip("/")
        return url_dictionary

    def parse_query_params(self, url):
        parsed = urlparse(url)
        params = dict(parse_qsl(parsed.query))
        return params

    def add_to_query_params(self, url, params: dict):
        if not params:
            return url
        current_params = self.parse_query_params(url)
        merged_params = urlencode({**current_params, **params})
        parsed = urlparse(url)
        parsed = parsed._replace(query=merged_params)
        return parsed.geturl()



from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_exception import FormRESTException, form_rest_exception
from pweb_form_rest.crud.pweb_response_maker import PWebHTTPResponse


class ValidationExceptionHandler:
    pweb_http_response: PWebHTTPResponse = PWebHTTPResponse()

    def exception_response(self, exception: Exception):
        if isinstance(exception, FormRESTException):
            return self.get_exception_response_object(exception)

    def get_exception_response_object(self, exception: FormRESTException):
        if exception.messageResponse:
            json_string_response = exception.messageResponse.to_dict()
        elif exception.error_details_exception:
            json_string_response = exception.errorResponse.to_dict()
        elif exception.message:
            json_string_response = self.get_rest_message_response(exception.message)
        else:
            json_string_response = self.get_rest_message_response(PWebFRConfig.UNKNOWN_ERROR)
        return self.pweb_http_response.json_string_response(json_string_response)

    def get_rest_message_response(self, message: str):
        error_exception = self.get_rest_error_exception(message)
        return error_exception.messageResponse.to_dict()

    def get_rest_error_exception(self, message: str):
        return form_rest_exception.error_message_exception(message)



import pweb_form_rest.common.pweb_fr_config
from ppy_common import ObjectHelper
from pweb_form_rest.common.pweb_fr_exception import FormRESTException
from pweb_form_rest.common.pweb_vexception_handler import ValidationExceptionHandler
from pweb_form_rest.swagger.pweb_swagger_ui import PWebSwaggerUI


class PWebFR:

    def init_app(self, pweb_app, config):
        self.update_config(config)
        self.register_swagger(pweb_app)
        self.register_validation_handler(pweb_app)

    def update_config(self, config):
        ObjectHelper.copy_config_property(config, pweb_form_rest.common.pweb_fr_config.PWebFRConfig)

    def register_swagger(self, pweb_app):
        pweb_swagger_ui = PWebSwaggerUI()
        pweb_swagger_ui.register(pweb_app)

    def register_validation_handler(self, pweb_app):
        if pweb_app:
            pweb_app.register_error_handler(FormRESTException, ValidationExceptionHandler().exception_response)



from pweb_form_rest.crud.pweb_crud_common import PWebCRUDCommon
from pweb_form_rest.crud.pweb_upload_processor import PWebUploadProcessor
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_orm import PWebBaseModel
from ppy_common import PyCommon


class FileDataCRUD(PWebCRUDCommon):
    pweb_upload_processor: PWebUploadProcessor = PWebUploadProcessor()

    def __init__(self, model: PWebBaseModel):
        self.model = model

    def upload_and_save_file(self, form_data, model, request_dto: PWebDataDTO, upload_path, override_names: dict = None):
        override_names = self.pweb_upload_processor.get_file_override_names(uuid=model.uuid, request_dto=request_dto, override_names=override_names)
        uploaded_file_names = self.pweb_upload_processor.validate_and_upload_multiple(form_data=form_data, request_dto=request_dto, upload_path=upload_path, override_name=override_names)
        for name in override_names:
            if hasattr(model, name) and name in uploaded_file_names:
                file_name = uploaded_file_names[name][override_names[name]]
                setattr(model, name, file_name)

        model = self.pweb_upload_processor.delete_orphan_files(override_names=override_names, form_data=form_data, model=model, upload_path=upload_path)
        model.save()
        return model

    def _process_data_and_file_upload(self, request_dto: PWebDataDTO, upload_path, override_names: dict = None, existing_model=None, form_data: dict = None):
        if not form_data:
            form_data = self.pweb_crud.get_form_data(request_dto)
        model = self.pweb_crud.populate_model(form_data, request_dto, instance=existing_model)
        if not model.uuid:
            model.uuid = PyCommon.uuid()
        return self.upload_and_save_file(form_data=form_data, model=model, request_dto=request_dto, upload_path=upload_path, override_names=override_names)

    def upload_file_data(self, request_dto: PWebDataDTO, upload_path, response_dto: PWebDataDTO = None, override_names: dict = None, response_message: str = "Successfully created!", form_data=None):
        model = self._process_data_and_file_upload(request_dto=request_dto, upload_path=upload_path, override_names=override_names, form_data=form_data)
        return self.message_or_data_response(model=model, response_dto=response_dto, response_message=response_message)

    def update_upload_file_data(self, request_dto: PWebDataDTO, upload_path, response_dto: PWebDataDTO = None, override_names: dict = None, response_message: str = "Successfully updated!", existing_model=None, form_data: dict = None, query=None):
        if not form_data:
            form_data = self.pweb_crud.get_form_data(data_dto=request_dto)

        if not existing_model:
            existing_model = self.get_by_id(form_data['id'], query=query, exception=True)

        model = self._process_data_and_file_upload(request_dto=request_dto, upload_path=upload_path, override_names=override_names, form_data=form_data, existing_model=existing_model)
        return self.message_or_data_response(model=model, response_dto=response_dto, response_message=response_message)

    def delete_file(self, upload_path, filename, response_message: str = "Successfully deleted!"):
        self.pweb_upload_processor.delete_file(upload_path=upload_path, filename=filename)
        return self.message_or_data_response(model=None, response_message=response_message)



from copy import copy
from flask import sessions
from marshmallow import EXCLUDE, ValidationError
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_exception import form_rest_exception
from pweb_form_rest.crud.pweb_request_data import RequestData
from pweb_form_rest.data.pweb_response_status import PWebResponseCode
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_orm import PWebBaseModel, and_, or_, pweb_orm


class PWebCRUD:
    request_data: RequestData = RequestData()

    def validate_data(self, data: dict, data_dto: PWebDataDTO, session=sessions, unknown=EXCLUDE):
        try:
            setattr(data_dto, "unknown", unknown)
            errors = data_dto.validate(data, session=session)
            if errors:
                raise form_rest_exception.process_validation_exception(errors=errors, message=PWebFRConfig.VALIDATION_ERROR)
            return data
        except ValidationError as error:
            raise form_rest_exception.process_validation_exception(errors=error.messages, message=PWebFRConfig.VALIDATION_ERROR)

    def load_model_from_dict(self, data: dict, data_dto: PWebDataDTO, session=sessions, instance=None):
        return data_dto.load(data, session=session, instance=instance, unknown=EXCLUDE)

    def load_allowed_data_from_dict(self, data: dict, data_dto: PWebDataDTO, instance=None):
        modified_data_dto = copy(data_dto)
        if hasattr(modified_data_dto, "_load_instance"):
            setattr(modified_data_dto, "_load_instance", False)
        return modified_data_dto.load(data, instance=instance, unknown=EXCLUDE)

    def populate_model(self, data: dict, data_dto: PWebDataDTO, session=sessions, instance=None):
        try:
            return self.load_model_from_dict(data, data_dto, session, instance)
        except ValidationError as error:
            raise form_rest_exception.process_validation_exception(errors=error.messages, message=PWebFRConfig.VALIDATION_ERROR)

    def get_json_data(self, data_dto: PWebDataDTO, is_validate=True, load_only=False, json_obj=None):
        if not json_obj:
            json_obj = self.request_data.json_data()
        else:
            if "data" not in json_obj:
                json_obj = {"data": json_obj}
        if not json_obj or "data" not in json_obj:
            raise form_rest_exception.error_message_exception(message=PWebFRConfig.INVALID_REQUEST_DATA, code=PWebResponseCode.error)
        json_obj = json_obj["data"]

        if is_validate:
            self.validate_data(json_obj, data_dto)

        if load_only:
            return self.load_allowed_data_from_dict(json_obj, data_dto)

        return json_obj

    def get_form_data(self, data_dto: PWebDataDTO, is_validate=True, is_populate_model=False, load_only=False, form_data=None):
        if not form_data:
            form_data = self.request_data.form_and_file_data()

        if not form_data:
            raise form_rest_exception.error_message_exception(message=PWebFRConfig.INVALID_REQUEST_DATA, code=PWebResponseCode.error)

        if is_validate:
            data = self.validate_data(form_data, data_dto)

            if is_populate_model:
                return self.load_model_from_dict(data, data_dto)

        if load_only:
            return self.load_allowed_data_from_dict(form_data, data_dto)

        return form_data

    def get_query_args(self, name, exception=True, exception_message=None, default=None, type=None):
        value = self.request_data.get_query_args_value(name, default=default, type=type)
        if not value and exception:
            if not exception_message:
                exception_message = name + " not found"
            raise form_rest_exception.error_message_exception(exception_message)
        return value

    def set_sort_order(self, model, query, default_field=PWebFRConfig.SORT_DEFAULT_FIELD_NAME, default_order=PWebFRConfig.SORT_DEFAULT_ORDER_NAME):
        sort_field = self.request_data.get_query_args_value(PWebFRConfig.SORT_FIELD_PARAM_NAME, default=default_field)
        sort_order = self.request_data.get_query_args_value(PWebFRConfig.SORT_ORDER_PARAM_NAME, default=default_order)
        if sort_order and (sort_order != "asc" and sort_order != "desc"):
            sort_order = default_order

        if not sort_order or not sort_field:
            return query

        if sort_order == "asc":
            return query.order_by(getattr(model, sort_field).asc())

        return query.order_by(getattr(model, sort_field).desc())

    def set_pagination(self, query, item_per_page=PWebFRConfig.TOTAL_ITEM_PER_PAGE):
        page: int = self.request_data.get_query_args_value(PWebFRConfig.PAGE_PARAM_NAME, default=0, type=int)
        per_page: int = self.request_data.get_query_args_value(PWebFRConfig.ITEM_PER_PAGE_PARAM_NAME, default=item_per_page, type=int)
        return query.paginate(page=page, per_page=per_page, error_out=False)

    def set_search(self, model, search_fields: list, query, search_text: str = None):
        like = []
        search = search_text
        if not search:
            search = self.request_data.get_query_args_value(PWebFRConfig.SEARCH_FIELD_PARAM_NAME)
        if search:
            for field in search_fields:
                like.append(getattr(model, field).ilike("%{}%".format(search)))
            if like:
                return query.filter(or_(*like))
        return query

    def list(self, model: PWebBaseModel, query=None, search_fields: list = None, enable_pagination: bool = True, enable_sort: bool = True,
             is_deleted=False, sort_field=None, sort_order=None, item_per_page=None, search_text: str = None
             ):
        if not query:
            query = model.query
        query = query.filter(getattr(model, "isDeleted") == is_deleted)

        if not sort_field:
            sort_field = PWebFRConfig.SORT_DEFAULT_FIELD_NAME

        if not sort_order:
            sort_order = PWebFRConfig.SORT_DEFAULT_ORDER_NAME

        if not item_per_page:
            item_per_page = PWebFRConfig.TOTAL_ITEM_PER_PAGE

        if enable_sort:
            query = self.set_sort_order(model, query=query, default_field=sort_field, default_order=sort_order)

        if search_fields:
            query = self.set_search(model, query=query, search_fields=search_fields, search_text=search_text)

        if enable_pagination:
            return self.set_pagination(query, item_per_page=item_per_page)

        return query.all()

    def get_by_id(self, model: PWebBaseModel, id: int, is_deleted: bool = False, exception: bool = False, message: str = "Entry Not Found!", query=None):
        if not query:
            query = model.query
        result = query.filter(and_(model.id == id, model.isDeleted == is_deleted)).first()
        if result:
            return result
        if not result and exception:
            raise form_rest_exception.error_message_exception(message)
        return None

    def get_by_ids(self, model: PWebBaseModel, ids, is_deleted: bool = False, exception: bool = False, message: str = "Not Found!", query=None):
        if not query:
            query = model.query
        result = query.filter(and_(model.id.in_(ids), model.isDeleted == is_deleted)).all()
        if result:
            return result
        if not result and exception:
            raise form_rest_exception.error_message_exception(message)
        return None

    def get_by_ids_not_in(self, model: PWebBaseModel, ids, is_deleted: bool = False, exception: bool = False, message: str = "Not Found!", query=None):
        if not query:
            query = model.query
        result = query.filter(and_(model.id.not_in(ids), model.isDeleted == is_deleted)).all()
        if result:
            return result
        if not result and exception:
            raise form_rest_exception.error_message_exception(message)
        return None

    def delete_all(self, model: PWebBaseModel, query=None):
        if not query:
            query = model.query
        query.delete()
        pweb_orm.session.commit()

    def delete_by_ids_not_in(self, model: PWebBaseModel, ids, query=None):
        if not query:
            query = model.query
        query.filter(and_(model.id.not_in(ids))).delete()
        pweb_orm.session.commit()

    def delete_by_ids_in(self, model: PWebBaseModel, ids, query=None):
        if not query:
            query = model.query
        query.filter(and_(model.id.in_(ids))).delete()
        pweb_orm.session.commit()

    def check_unique(self, model: PWebBaseModel, field: str, value, model_id=None, exception: bool = True, message: str = "Already used", query=None):
        if not query:
            query = model.query
        query = query.filter(getattr(model, field) == value)
        if model_id:
            query = query.filter(model.id != model_id)
        result = query.first()
        if result and exception:
            raise form_rest_exception.error_details_exception("Unique filed error", details={field: message})




import types
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.common.pweb_fr_constant import SwaggerCommon
from pweb_form_rest.swagger.pweb_swagger_param_def import SwaggerParamDef


class PWebSDProcessor:
    _pweb_app = None
    _ignore_verbs = {"HEAD", "OPTIONS"}
    _swagger_data_type_mapping = {
        "string": "string",
        "int": "integer",
        "float": "number",
        "path": "string",
        "any": "string",
        "uuid": "string"
    }

    def __init__(self, pweb_app):
        self._pweb_app = pweb_app

    def get_action_definitions(self):
        definitions = []
        for rule in self._pweb_app.url_map.iter_rules():
            endpoint = self._pweb_app.view_functions[rule.endpoint]
            if isinstance(endpoint, types.FunctionType):
                function_name = endpoint.__name__
                if function_name and self._is_it_swagger_definition(endpoint):
                    definition = endpoint(pass_definition=True)
                    definition = self._process_action_decorator(definition, rule)
                    if definition:
                        definitions.append(definition)
        return definitions

    def _is_it_swagger_definition(self, endpoint) -> bool:
        try:
            name = endpoint.__pweb_swagger__
            if name == SwaggerCommon.PWEB_SWAGGER:
                return True
        except:
            return False
        return False

    def _process_action_decorator(self, definition: SwaggerParamDef, rule):
        definition = self._get_url_params(definition, rule)
        definition.methods = self._get_action_methods(rule, definition)
        definition = self._get_default_tag_name(definition, rule)
        return definition

    def _get_url_params(self, definition: SwaggerParamDef, rule) -> SwaggerParamDef:
        path = []
        url_and_data_type = self._extract_url_to_params(rule.rule)
        data_type = url_and_data_type['url_map']
        for param in rule.arguments:
            if param in data_type:
                path.append((param, data_type[param], True))
        definition.url = url_and_data_type['url']
        if definition.url_params and isinstance(definition.url_params, list):
            definition.url_params.extend(path)
        else:
            definition.url_params = path
        return definition

    def _extract_url_to_params(self, url):
        url_map = {}
        if url:
            fragments = url.split("/")
            for fragment in fragments:
                if fragment and fragment.startswith("<"):
                    actual_fragment = fragment
                    type_input = fragment.replace("<", "").replace(">", "").split(":")
                    if len(type_input) == 2 and (type_input[0] in self._swagger_data_type_mapping):
                        param_name = type_input[1]
                        url_map[param_name] = self._swagger_data_type_mapping[type_input[0]]
                    else:
                        url_map[type_input[0]] = "string"
                        param_name = type_input[0]
                    url = url.replace(actual_fragment, "{" + param_name + "}")
        return {"url": url, "url_map": url_map}

    def _get_action_methods(self, rule, definition: SwaggerParamDef):
        methods = []
        if definition.method:
            methods.append(definition.method)
            return methods
        for method in rule.methods.difference(self._ignore_verbs):
            methods.append(method)
        return methods

    def _get_default_tag_name(self, definition: SwaggerParamDef, rule):
        if definition.tag:
            definition.tags.append(definition.tag)
            return definition
        endpoint_name = rule.endpoint
        end = endpoint_name.find(".")
        total = len(endpoint_name)
        definition.tags = []
        if end != -1 and total > end:
            endpoint_name = endpoint_name[0:end]
            endpoint_name = endpoint_name.replace("_", " ")
            endpoint_name = endpoint_name.title()
            definition.tags.append(endpoint_name)
        else:
            definition.tags.append(PWebFRConfig.SWAGGER_DEFAULT_TAG_NAME)
        return definition



class PWebResponseStatus:
    success = "success"
    error = "error"


class PWebResponseCode:
    # SUCCESS Codes
    success = 2200

    # ERROR Codes
    error = 5100
    validation_error = 5101

    invalid_token_code = 5500
    token_expired_code = 5501
    token_error_code = 5502


class PWebHTTPCode:
    # SUCCESS Codes
    OK = 200



import os
from werkzeug.datastructures import FileStorage
from ppy_file_text import StringUtil


class PWebWebFileUtil:

    @staticmethod
    def get_file_size(file_object: FileStorage, default=0):
        if file_object.content_length:
            return file_object.content_length

        try:
            current_position = file_object.tell()
            file_object.seek(0, 2)
            size = file_object.tell()
            file_object.seek(current_position)
            return size
        except (AttributeError, IOError):
            pass

        return default

    @staticmethod
    def is_valid_file_size(file_object: FileStorage, max_size_kb):
        size = PWebWebFileUtil.get_file_size(file_object)
        size_in_kb = round(size / 1024, 4)
        if size_in_kb <= max_size_kb:
            return True
        return False

    @staticmethod
    def allowed_file(filename, file_extensions: list):
        return '.' in filename and filename.rsplit('.', 1)[1].lower() in file_extensions

    @staticmethod
    def process_file_name(file_name: str):
        if file_name:
            for sep in os.path.sep, os.path.altsep:
                if sep:
                    file_name = file_name.replace(sep, " ")

            file_name = StringUtil.camelcase_to(file_name, "-")
            file_name = StringUtil.find_and_replace_with(file_name, "_", "-")
            file_name = StringUtil.replace_space_with(file_name, "-")
            file_name = file_name.lower()
        return file_name

    @staticmethod
    def get_file_extension(filename):
        if '.' in filename:
            return filename.rsplit('.', 1)[1].lower()



import json
from flask import get_flashed_messages, render_template, render_template_string
from ppy_file_text import TextFileMan
from pweb_form_rest.common.pweb_fr_config import PWebFRConfig
from pweb_form_rest.crud.pweb_request_data import RequestData
from pweb_form_rest.form.pweb_form import PWebForm


class UIUtil:
    request_data: RequestData = RequestData()

    @property
    def pweb_flashed_messages(self):
        messages = get_flashed_messages(with_categories=True)
        response = {}
        for key, value in messages:
            response[key] = value
        return json.dumps(response)

    def base_url(self):
        info = self.request_data.get_url_info()
        return info.baseURL

    def set_base_js(self, obj_prefix="PWeb"):
        script = f"""
        <script>
            {obj_prefix}.baseURL = "{self.base_url()}"
        </script>
        """
        return script


class PWebSSRUIHelper:

    def get_helper(self) -> dict:
        return {}

    def _process_render_params(self, params: dict = None, form: PWebForm = None):
        if not params:
            params = {}

        if form and form.definition:
            params["form"] = form.definition

        helper = self.get_helper()
        if helper:
            params.update(helper)
        params["util"] = UIUtil()
        return params

    def render_html_file(self, file_path, params: dict = None, form: PWebForm = None, is_exception: bool = False):
        html = TextFileMan.get_text_from_file(file_path=file_path, is_exception=is_exception, exception_message="File Not found!", default="")
        params = self._process_render_params(params=params, form=form)
        return render_template_string(html, **params)

    def render(self, view_name, params: dict = None, form: PWebForm = None):
        params = self._process_render_params(params=params, form=form)
        return render_template(f"{view_name}.html", **params)


def _get_ssr_ui_helper(ssr_ui_helper: PWebSSRUIHelper = None):
    if not ssr_ui_helper:
        ssr_ui_helper = PWebSSRUIHelper()
        if PWebFRConfig.SSR_UI_HELPER and isinstance(PWebFRConfig.SSR_UI_HELPER, PWebSSRUIHelper):
            ssr_ui_helper = PWebFRConfig.SSR_UI_HELPER
    return ssr_ui_helper


def ssr_ui_render(view_name, params: dict = None, form: PWebForm = None, ssr_ui_helper: PWebSSRUIHelper = None):
    ssr_ui_helper = _get_ssr_ui_helper(ssr_ui_helper=ssr_ui_helper)
    return ssr_ui_helper.render(view_name=view_name, params=params, form=form)


def ssr_ui_render_html_file(file_path, params: dict = None, form: PWebForm = None, ssr_ui_helper: PWebSSRUIHelper = None, is_exception: bool = False):
    ssr_ui_helper = _get_ssr_ui_helper(ssr_ui_helper=ssr_ui_helper)
    return ssr_ui_helper.render_html_file(file_path=file_path, params=params, form=form, is_exception=is_exception)



import json
import typing as t
from flask import make_response, send_file
from pweb_form_rest.data.pweb_response_data import PWebMessageResponse, PWebDataResponse, PWebPagination, \
    PWebPaginateResponse
from pweb_form_rest.data.pweb_response_status import PWebResponseCode, PWebResponseStatus
from pweb_form_rest.schema.pweb_rest_schema import PWebDataDTO
from pweb_orm import PWebBaseModel


class PWebHTTPResponse:

    def response(self, response_data, code=200, headers: dict = None):
        response_obj = make_response(response_data, code)
        if headers:
            response_obj.headers.update(headers)
        return response_obj

    def json_string_response(self, json_string: str, code=200, headers: dict = None):
        _headers = {
            "Content-Type": "application/json; charset=utf-8",
        }
        if headers and isinstance(headers, dict):
            _headers.update(headers)
        return self.response(json_string, code=code, headers=_headers)

    def json_response(self, dictionary, code=200, headers: dict = None):
        json_object = {}
        if dictionary:
            json_object = dictionary
        response_string = "{}"
        try:
            response_string = json.dumps(json_object)
        except Exception as e:
            response_string = "{}"
        return self.json_string_response(response_string, code, headers)

    def file_response(self, file_path, download_name: t.Optional[str] = None):
        return send_file(file_path, download_name=download_name)

    def text_response(self, text: str, code=200, headers: dict = None):
        return self.response(text, code=code, headers=headers)


class ResponseMaker:
    headers: dict = None
    http_response: PWebHTTPResponse = PWebHTTPResponse()

    def add_header(self, key: str, value):
        if not self.headers:
            self.headers = {}
        self.headers[key] = value

    def message(self, message: str, status: str, code: str, http_code=200):
        message_response = PWebMessageResponse()
        message_response.message = message
        message_response.status = status
        message_response.code = code
        return self.http_response.json_string_response(message_response.to_dict(), http_code, self.headers)

    def success_message(self, message: str, code: str = PWebResponseCode.success, http_code=200):
        return self.message(message, PWebResponseStatus.success, code, http_code)

    def error_message(self, message: str, code: str = PWebResponseCode.error, http_code=200):
        return self.message(message, PWebResponseStatus.error, code, http_code)

    def data_type_response(self, model: PWebBaseModel, response_dto: PWebDataDTO, many=False, status: str = PWebResponseStatus.success, code: str = PWebResponseCode.success):
        data_response = PWebDataResponse()
        data_response.status = status
        data_response.code = code
        data_response.add_data(model, response_dto, many)
        return data_response

    def data_response(self, model: PWebBaseModel, response_dto: PWebDataDTO, many=False, status: str = PWebResponseStatus.success, code: str = PWebResponseCode.success, http_code=200):
        data_response = self.data_type_response(model=model, response_dto=response_dto, many=many, status=status, code=code)
        return self.http_response.json_string_response(data_response.to_dict(), http_code, self.headers)

    def set_pagination_data(self, model: PWebBaseModel):
        pagination = PWebPagination()
        pagination.page = model.page
        pagination.totalPage = model.pages
        pagination.itemPerPage = model.per_page
        pagination.total = model.total
        return pagination

    def _get_paginate_response_object(self,  model: PWebBaseModel):
        pagination = self.set_pagination_data(model)
        response = PWebPaginateResponse()
        response.status = PWebResponseStatus.success
        response.code = PWebResponseCode.success
        response.pagination = pagination
        return response

    def paginate_type_response(self, model: PWebBaseModel, response_dto: PWebDataDTO):
        response = self._get_paginate_response_object(model=model)
        response.data = model.items
        response.add_only_data(model.items, response_dto, True)
        return response.to_dict()

    def paginate_response(self, model: PWebBaseModel, response_dto: PWebDataDTO):
        response_dict = self.paginate_type_response(model, response_dto)
        return self.http_response.json_string_response(response_dict, headers=self.headers)

    def list_data_type_response(self, model: PWebBaseModel, response_dto: PWebDataDTO, status: str = PWebResponseStatus.success, code: str = PWebResponseCode.success, http_code=200):
        data_response = PWebDataResponse()
        data_response.status = status
        data_response.code = code
        data_response.add_data(model, response_dto, True)
        return self.http_response.json_string_response(data_response.to_dict(), http_code, self.headers)

    def dictionary_object_response(self, data: dict, status: str = PWebResponseStatus.success, code: str = PWebResponseCode.success, http_code=200, message=None):
        data_response = PWebDataResponse()
        data_response.status = status
        data_response.code = code
        data_response.data = data
        data_response.message = message
        return self.http_response.json_string_response(data_response.to_dict(), http_code, self.headers)

    def list_object_response(self, data: list, status: str = PWebResponseStatus.success, code: str = PWebResponseCode.success, http_code=200):
        data_response = PWebDataResponse()
        data_response.status = status
        data_response.code = code
        data_response.data = data
        data_response.add_data(None, None, True)
        return self.http_response.json_string_response(data_response.to_dict(), http_code, self.headers)
